# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from uuid import NAMESPACE_OID, UUID, uuid5

NIL_UUID = UUID(int=0)


def generate_uuid(*values: str | None, namespace: UUID = NAMESPACE_OID) -> str:
    """
    Convert input values to deterministic UUID string representation.

    This function is only intended to generate a hash which used as an identifier, not for any security use.

    Generates a UUID v5 (SHA-1 + Namespace) for each value provided,
    and this UUID is used as the Namespace for the next element.

    If only one non-None value is provided to the function, then the result of the function
    would be the same as result of ``uuid.uuid5``.

    All ``None`` values are replaced by NIL UUID.  If it only one value is provided then return NIL UUID.

    :param namespace: Initial namespace value to pass into the ``uuid.uuid5`` function.
    """
    if not values:
        raise ValueError("Expected at least 1 argument")

    if len(values) == 1 and values[0] is None:
        return str(NIL_UUID)

    result = namespace
    for item in values:
        result = uuid5(result, item if item is not None else str(NIL_UUID))

    return str(result)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import Mock, patch

import pytest

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if not AIRFLOW_V_3_0_PLUS:
    pytest.skip("AWS auth manager is only compatible with Airflow >= 3.0.0", allow_module_level=True)

from fastapi.testclient import TestClient

from airflow.api_fastapi.app import AUTH_MANAGER_FASTAPI_APP_PREFIX, create_app

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.mock_plugins import mock_plugin_manager

# onelogin is optional dependency from apache-airflow-providers-amazon[python3-saml]
# we want to skip it for the lowest dependency checks as it does not install extra dependencies
# https://github.com/apache/airflow/pull/50449#issuecomment-2897572327
OneLogin_Saml2_IdPMetadataParser = pytest.importorskip(
    "onelogin.saml2.idp_metadata_parser"
).OneLogin_Saml2_IdPMetadataParser

SAML_METADATA_URL = "/saml/metadata"
SAML_METADATA_PARSED = {
    "idp": {
        "entityId": "https://portal.sso.us-east-1.amazonaws.com/saml/assertion/<assertion>",
        "singleSignOnService": {
            "url": "https://portal.sso.us-east-1.amazonaws.com/saml/assertion/<assertion>",
            "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        },
        "singleLogoutService": {
            "url": "https://portal.sso.us-east-1.amazonaws.com/saml/logout/<assertion>",
            "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        },
        "x509cert": "<cert>",
    },
    "security": {"authnRequestsSigned": False},
    "sp": {"NameIDFormat": "urn:oasis:names:tc:SAML:2.0:nameid-format:transient"},
}


@pytest.fixture
def test_client():
    with conf_vars(
        {
            (
                "core",
                "auth_manager",
            ): "airflow.providers.amazon.aws.auth_manager.aws_auth_manager.AwsAuthManager",
            ("aws_auth_manager", "saml_metadata_url"): SAML_METADATA_URL,
        }
    ):
        with (
            patch.object(OneLogin_Saml2_IdPMetadataParser, "parse_remote") as mock_parse_remote,
            patch(
                "airflow.providers.amazon.aws.auth_manager.avp.facade.AwsAuthManagerAmazonVerifiedPermissionsFacade.is_policy_store_schema_up_to_date"
            ) as mock_is_policy_store_schema_up_to_date,
        ):
            mock_is_policy_store_schema_up_to_date.return_value = True
            mock_parse_remote.return_value = SAML_METADATA_PARSED
            yield TestClient(create_app())


def get_login_callback_response(relay_state: str):
    with conf_vars(
        {
            (
                "core",
                "auth_manager",
            ): "airflow.providers.amazon.aws.auth_manager.aws_auth_manager.AwsAuthManager",
            ("aws_auth_manager", "saml_metadata_url"): SAML_METADATA_URL,
            ("api", "ssl_cert"): "false",
        }
    ):
        with (
            patch.object(OneLogin_Saml2_IdPMetadataParser, "parse_remote") as mock_parse_remote,
            patch(
                "airflow.providers.amazon.aws.auth_manager.routes.login._init_saml_auth"
            ) as mock_init_saml_auth,
            patch(
                "airflow.providers.amazon.aws.auth_manager.avp.facade.AwsAuthManagerAmazonVerifiedPermissionsFacade.is_policy_store_schema_up_to_date"
            ) as mock_is_policy_store_schema_up_to_date,
        ):
            mock_is_policy_store_schema_up_to_date.return_value = True
            mock_parse_remote.return_value = SAML_METADATA_PARSED

            auth = Mock()
            auth.is_authenticated.return_value = True
            auth.get_nameid.return_value = "user_id"
            auth.get_attributes.return_value = {
                "id": ["1"],
                "groups": ["group_1", "group_2"],
                "email": ["email"],
            }
            mock_init_saml_auth.return_value = auth
            client = TestClient(create_app())
            return client.post(
                AUTH_MANAGER_FASTAPI_APP_PREFIX + "/login_callback",
                follow_redirects=False,
                data={"RelayState": relay_state},
            )


@mock_plugin_manager(plugins=[])
class TestLoginRouter:
    @pytest.mark.parametrize(
        "url",
        ["/login", "/login/token"],
    )
    def test_login(self, test_client, url):
        response = test_client.get(AUTH_MANAGER_FASTAPI_APP_PREFIX + url, follow_redirects=False)
        assert response.status_code == 307
        assert "location" in response.headers
        assert response.headers["location"].startswith(
            "https://portal.sso.us-east-1.amazonaws.com/saml/assertion/"
        )

    def test_login_callback_successful_with_relay_state_redirect(self):
        response = get_login_callback_response("login-redirect")
        assert response.status_code == 303
        assert "location" in response.headers
        assert "_token" in response.cookies
        assert response.headers["location"].startswith("http://localhost:8080/")

    def test_login_callback_successful_with_relay_state_token(self):
        response = get_login_callback_response("login-token")
        assert response.status_code == 200
        assert "access_token" in response.json()

    def test_login_callback_with_invalid_relay_state(self):
        response = get_login_callback_response("dummy")
        assert response.status_code == 500

    def test_login_callback_unsuccessful(self):
        with conf_vars(
            {
                (
                    "core",
                    "auth_manager",
                ): "airflow.providers.amazon.aws.auth_manager.aws_auth_manager.AwsAuthManager",
                ("aws_auth_manager", "saml_metadata_url"): SAML_METADATA_URL,
            }
        ):
            with (
                patch.object(OneLogin_Saml2_IdPMetadataParser, "parse_remote") as mock_parse_remote,
                patch(
                    "airflow.providers.amazon.aws.auth_manager.routes.login._init_saml_auth"
                ) as mock_init_saml_auth,
                patch(
                    "airflow.providers.amazon.aws.auth_manager.avp.facade.AwsAuthManagerAmazonVerifiedPermissionsFacade.is_policy_store_schema_up_to_date"
                ) as mock_is_policy_store_schema_up_to_date,
            ):
                mock_is_policy_store_schema_up_to_date.return_value = True
                mock_parse_remote.return_value = SAML_METADATA_PARSED

                auth = Mock()
                auth.is_authenticated.return_value = False
                mock_init_saml_auth.return_value = auth
                client = TestClient(create_app())
                response = client.post(AUTH_MANAGER_FASTAPI_APP_PREFIX + "/login_callback")
                assert response.status_code == 500



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import os
from unittest import mock

import pytest
from moto import mock_aws

from airflow.models import DAG, DagRun, TaskInstance
from airflow.providers.amazon.aws.hooks.athena import AthenaHook
from airflow.providers.amazon.aws.operators.athena import AthenaOperator
from airflow.providers.amazon.aws.triggers.athena import AthenaTrigger
from airflow.providers.common.compat.openlineage.facet import (
    Dataset,
    ExternalQueryRunFacet,
    Identifier,
    SchemaDatasetFacet,
    SchemaDatasetFacetFields,
    SQLJobFacet,
    SymlinksDatasetFacet,
)
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred
from airflow.providers.openlineage.extractors import OperatorLineage
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType

from tests_common.test_utils.compat import timezone
from tests_common.test_utils.dag import sync_dag_to_db
from tests_common.test_utils.taskinstance import create_task_instance, get_template_context
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS
from unit.amazon.aws.utils.test_template_fields import validate_template_fields

TEST_DAG_ID = "unit_tests"
DEFAULT_DATE = timezone.datetime(2018, 1, 1)
ATHENA_QUERY_ID = "eac29bf8-daa1-4ffc-b19a-0db31dc3b784"

MOCK_DATA = {
    "task_id": "test_athena_operator",
    "query": "SELECT * FROM TEST_TABLE",
    "database": "TEST_DATABASE",
    "catalog": "AwsDataCatalog",
    "outputLocation": "s3://test_s3_bucket/",
    "client_request_token": "eac427d0-1c6d-4dfb-96aa-2835d3ac6595",
    "workgroup": "primary",
}

query_context = {"Database": MOCK_DATA["database"], "Catalog": MOCK_DATA["catalog"]}
result_configuration = {"OutputLocation": MOCK_DATA["outputLocation"]}


@mock_aws
class TestAthenaOperator:
    @pytest.fixture(autouse=True)
    def _setup_test_cases(self):
        args = {
            "owner": "airflow",
            "start_date": DEFAULT_DATE,
        }

        self.dag = DAG(TEST_DAG_ID, default_args=args, schedule="@once")

        self.default_op_kwargs = dict(
            task_id="test_athena_operator",
            query="SELECT * FROM TEST_TABLE",
            database="TEST_DATABASE",
            client_request_token="eac427d0-1c6d-4dfb-96aa-2835d3ac6595",
            sleep_time=0,
            max_polling_attempts=3,
        )
        self.athena = AthenaOperator(
            **self.default_op_kwargs, output_location="s3://test_s3_bucket/", aws_conn_id=None, dag=self.dag
        )

        with mock.patch("airflow.providers.amazon.aws.links.athena.AthenaQueryResultsLink.persist") as m:
            self.mocked_athena_result_link = m
            yield

    def test_base_aws_op_attributes(self):
        op = AthenaOperator(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None
        assert op.hook.log_query is True

        op = AthenaOperator(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
            log_query=False,
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42
        assert op.hook.log_query is False

    def test_init(self):
        assert self.athena.task_id == MOCK_DATA["task_id"]
        assert self.athena.query == MOCK_DATA["query"]
        assert self.athena.database == MOCK_DATA["database"]
        assert self.athena.catalog == MOCK_DATA["catalog"]
        assert self.athena.client_request_token == MOCK_DATA["client_request_token"]
        assert self.athena.sleep_time == 0

    @mock.patch.object(AthenaHook, "check_query_status", side_effect=("SUCCEEDED",))
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_hook_run_override_catalog(self, mock_conn, mock_run_query, mock_check_query_status):
        query_context_catalog = {"Database": MOCK_DATA["database"], "Catalog": "MyCatalog"}
        self.athena.catalog = "MyCatalog"
        self.athena.execute({})
        mock_run_query.assert_called_once_with(
            MOCK_DATA["query"],
            query_context_catalog,
            result_configuration,
            MOCK_DATA["client_request_token"],
            MOCK_DATA["workgroup"],
        )
        assert mock_check_query_status.call_count == 1

    @mock.patch.object(AthenaHook, "check_query_status", side_effect=("SUCCEEDED",))
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_hook_run_small_success_query(self, mock_conn, mock_run_query, mock_check_query_status):
        self.athena.execute({})
        mock_run_query.assert_called_once_with(
            MOCK_DATA["query"],
            query_context,
            result_configuration,
            MOCK_DATA["client_request_token"],
            MOCK_DATA["workgroup"],
        )
        assert mock_check_query_status.call_count == 1

        # Validate call persist Athena Query result link
        self.mocked_athena_result_link.assert_called_once_with(
            aws_partition=mock.ANY,
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            query_execution_id=ATHENA_QUERY_ID,
        )

    @mock.patch.object(
        AthenaHook,
        "check_query_status",
        side_effect="SUCCEEDED",
    )
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_hook_run_big_success_query(self, mock_conn, mock_run_query, mock_check_query_status):
        self.athena.execute({})
        mock_run_query.assert_called_once_with(
            MOCK_DATA["query"],
            query_context,
            result_configuration,
            MOCK_DATA["client_request_token"],
            MOCK_DATA["workgroup"],
        )

    @mock.patch.object(AthenaHook, "get_state_change_reason")
    @mock.patch.object(AthenaHook, "check_query_status", return_value="FAILED")
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_hook_run_failure_query(
        self,
        mock_conn,
        mock_run_query,
        mock_check_query_status,
        mock_get_state_change_reason,
    ):
        with pytest.raises(AirflowException):
            self.athena.execute({})
        mock_run_query.assert_called_once_with(
            MOCK_DATA["query"],
            query_context,
            result_configuration,
            MOCK_DATA["client_request_token"],
            MOCK_DATA["workgroup"],
        )
        assert mock_get_state_change_reason.call_count == 1

    @mock.patch.object(AthenaHook, "check_query_status", return_value="CANCELLED")
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_hook_run_cancelled_query(self, mock_conn, mock_run_query, mock_check_query_status):
        with pytest.raises(AirflowException):
            self.athena.execute({})
        mock_run_query.assert_called_once_with(
            MOCK_DATA["query"],
            query_context,
            result_configuration,
            MOCK_DATA["client_request_token"],
            MOCK_DATA["workgroup"],
        )

    @mock.patch.object(AthenaHook, "check_query_status", return_value="RUNNING")
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_hook_run_failed_query_with_max_tries(self, mock_conn, mock_run_query, mock_check_query_status):
        with pytest.raises(AirflowException):
            self.athena.execute({})
        mock_run_query.assert_called_once_with(
            MOCK_DATA["query"],
            query_context,
            result_configuration,
            MOCK_DATA["client_request_token"],
            MOCK_DATA["workgroup"],
        )

    @pytest.mark.db_test
    @mock.patch.object(AthenaHook, "check_query_status", side_effect=("SUCCEEDED",))
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_return_value(
        self,
        mock_conn,
        mock_run_query,
        mock_check_query_status,
        session,
        clean_dags_dagruns_and_dagbundles,
        testing_dag_bundle,
    ):
        """Test we return the right value -- that will get put in to XCom by the execution engine"""
        if AIRFLOW_V_3_0_PLUS:
            from airflow.models.dag_version import DagVersion

            sync_dag_to_db(self.dag)
            dag_version = DagVersion.get_latest_version(self.dag.dag_id)
            ti = create_task_instance(task=self.athena, dag_version_id=dag_version.id)
            dag_run = DagRun(
                dag_id=self.dag.dag_id,
                logical_date=timezone.utcnow(),
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
        else:
            dag_run = DagRun(
                dag_id=self.dag.dag_id,
                execution_date=timezone.utcnow(),
                run_id="test",
                run_type=DagRunType.MANUAL,
                state=DagRunState.RUNNING,
            )
            ti = TaskInstance(task=self.athena)
        ti.dag_run = dag_run
        session.add(ti)
        session.commit()
        assert self.athena.execute(get_template_context(ti, self.athena)) == ATHENA_QUERY_ID

    @mock.patch.object(AthenaHook, "check_query_status", side_effect=("SUCCEEDED",))
    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_optional_output_location(self, mock_conn, mock_run_query, mock_check_query_status):
        op = AthenaOperator(**self.default_op_kwargs, aws_conn_id=None)

        op.execute({})
        mock_run_query.assert_called_once_with(
            MOCK_DATA["query"],
            query_context,
            {},  # Should be an empty dict since we do not provide output_location
            MOCK_DATA["client_request_token"],
            MOCK_DATA["workgroup"],
        )

    @mock.patch.object(AthenaHook, "run_query", return_value=ATHENA_QUERY_ID)
    def test_is_deferred(self, mock_run_query):
        self.athena.deferrable = True

        with pytest.raises(TaskDeferred) as deferred:
            self.athena.execute(None)

        assert isinstance(deferred.value.trigger, AthenaTrigger)

        # Validate call persist Athena Query result link
        self.mocked_athena_result_link.assert_called_once_with(
            aws_partition=mock.ANY,
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            query_execution_id=ATHENA_QUERY_ID,
        )

    def test_execute_complete_reassigns_query_execution_id_after_deferring(self):
        """Assert that we use query_execution_id from event after deferral."""

        operator = AthenaOperator(
            task_id="test_athena_operator",
            query="SELECT * FROM TEST_TABLE",
            database="TEST_DATABASE",
            deferrable=True,
        )
        assert operator.query_execution_id is None

        query_execution_id = "123456"
        operator.execute_complete(
            context=None,
            event={"status": "success", "value": query_execution_id},
        )
        assert operator.query_execution_id == query_execution_id

    @mock.patch.object(AthenaHook, "region_name", new_callable=mock.PropertyMock)
    @mock.patch.object(AthenaHook, "get_conn")
    def test_operator_openlineage_data(self, mock_conn, mock_region_name):
        mock_region_name.return_value = "eu-west-1"

        def mock_get_table_metadata(CatalogName, DatabaseName, TableName):
            with open(os.path.dirname(__file__) + "/athena_metadata.json") as f:
                return json.load(f)[TableName]

        mock_conn.return_value.get_table_metadata = mock_get_table_metadata

        op = AthenaOperator(
            task_id="test_athena_openlineage",
            query="INSERT INTO TEST_TABLE SELECT CUSTOMER_EMAIL FROM DISCOUNTS",
            database="TEST_DATABASE",
            output_location="s3://test_s3_bucket",
            client_request_token="eac427d0-1c6d-4dfb-96aa-2835d3ac6595",
            sleep_time=0,
            max_polling_attempts=3,
            dag=self.dag,
        )
        op.query_execution_id = "12345"  # Mocking what will be available after execution

        expected_lineage = OperatorLineage(
            inputs=[
                Dataset(
                    namespace="awsathena://athena.eu-west-1.amazonaws.com",
                    name="AwsDataCatalog.TEST_DATABASE.DISCOUNTS",
                    facets={
                        "symlinks": SymlinksDatasetFacet(
                            identifiers=[
                                Identifier(
                                    namespace="s3://bucket",
                                    name="/discount/data/path/",
                                    type="TABLE",
                                )
                            ],
                        ),
                        "schema": SchemaDatasetFacet(
                            fields=[
                                SchemaDatasetFacetFields(
                                    name="ID",
                                    type="int",
                                    description="from deserializer",
                                ),
                                SchemaDatasetFacetFields(
                                    name="AMOUNT_OFF",
                                    type="int",
                                    description="from deserializer",
                                ),
                                SchemaDatasetFacetFields(
                                    name="CUSTOMER_EMAIL",
                                    type="varchar",
                                    description="from deserializer",
                                ),
                                SchemaDatasetFacetFields(
                                    name="STARTS_ON",
                                    type="timestamp",
                                    description="from deserializer",
                                ),
                                SchemaDatasetFacetFields(
                                    name="ENDS_ON",
                                    type="timestamp",
                                    description=None,
                                ),
                            ],
                        ),
                    },
                )
            ],
            outputs=[
                Dataset(
                    namespace="awsathena://athena.eu-west-1.amazonaws.com",
                    name="AwsDataCatalog.TEST_DATABASE.TEST_TABLE",
                    facets={
                        "symlinks": SymlinksDatasetFacet(
                            identifiers=[
                                Identifier(
                                    namespace="s3://bucket",
                                    name="/data/test_table/data/path",
                                    type="TABLE",
                                )
                            ],
                        ),
                        "schema": SchemaDatasetFacet(
                            fields=[
                                SchemaDatasetFacetFields(
                                    name="column",
                                    type="string",
                                    description="from deserializer",
                                )
                            ],
                        ),
                    },
                ),
                Dataset(namespace="s3://test_s3_bucket", name="/"),
            ],
            job_facets={
                "sql": SQLJobFacet(
                    query="INSERT INTO TEST_TABLE SELECT CUSTOMER_EMAIL FROM DISCOUNTS",
                )
            },
            run_facets={"externalQuery": ExternalQueryRunFacet(externalQueryId="12345", source="awsathena")},
        )
        assert op.get_openlineage_facets_on_complete(None) == expected_lineage

    def test_template_fields(self):
        validate_template_fields(self.athena)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.amazon.aws.links.base_aws import BASE_AWS_CONSOLE_LINK, BaseAwsLink


class EC2InstanceLink(BaseAwsLink):
    """Helper class for constructing Amazon EC2 instance links."""

    name = "Instance"
    key = "_instance_id"
    format_str = (
        BASE_AWS_CONSOLE_LINK + "/ec2/home?region={region_name}#InstanceDetails:instanceId={instance_id}"
    )


class EC2InstanceDashboardLink(BaseAwsLink):
    """
    Helper class for constructing Amazon EC2 console links.

    This is useful for displaying the list of EC2 instances, rather
    than a single instance.
    """

    name = "EC2 Instances"
    key = "_instance_dashboard"
    format_str = BASE_AWS_CONSOLE_LINK + "/ec2/home?region={region_name}#Instances:instanceId=:{instance_ids}"

    @staticmethod
    def format_instance_id_filter(instance_ids: list[str]) -> str:
        return ",:".join(instance_ids)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.opensearch_serverless import OpenSearchServerlessHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.opensearch_serverless import (
    OpenSearchServerlessCollectionActiveTrigger,
)
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf
from airflow.utils.helpers import exactly_one

if TYPE_CHECKING:
    from airflow.sdk import Context


class OpenSearchServerlessCollectionActiveSensor(AwsBaseSensor[OpenSearchServerlessHook]):
    """
    Poll the state of the Collection until it reaches a terminal state; fails if the query fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:OpenSearchServerlessCollectionAvailableSensor`

    :param collection_id: A collection ID. You can't provide a name and an ID in the same request.
    :param collection_name: A collection name. You can't provide a name and an ID in the same request.

    :param deferrable: If True, the sensor will operate in deferrable more. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 10)
    :param max_retries: Number of times before returning the current state (default: 60)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES = ("CREATING",)
    FAILURE_STATES = (
        "DELETING",
        "FAILED",
    )
    SUCCESS_STATES = ("ACTIVE",)
    FAILURE_MESSAGE = "OpenSearch Serverless Collection sensor failed"

    aws_hook_class = OpenSearchServerlessHook
    template_fields: Sequence[str] = aws_template_fields(
        "collection_id",
        "collection_name",
    )
    ui_color = "#66c3ff"

    def __init__(
        self,
        *,
        collection_id: str | None = None,
        collection_name: str | None = None,
        poke_interval: int = 10,
        max_retries: int = 60,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        if not exactly_one(collection_id is None, collection_name is None):
            raise AttributeError("Either collection_ids or collection_names must be provided, not both.")
        self.collection_id = collection_id
        self.collection_name = collection_name

        self.poke_interval = poke_interval
        self.max_retries = max_retries
        self.deferrable = deferrable

    def poke(self, context: Context, **kwargs) -> bool:
        call_args = (
            {"ids": [str(self.collection_id)]}
            if self.collection_id
            else {"names": [str(self.collection_name)]}
        )
        state = self.hook.conn.batch_get_collection(**call_args)["collectionDetails"][0]["status"]

        if state in self.FAILURE_STATES:
            raise AirflowException(self.FAILURE_MESSAGE)

        if state in self.INTERMEDIATE_STATES:
            return False
        return True

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=OpenSearchServerlessCollectionActiveTrigger(
                    collection_id=self.collection_id,
                    collection_name=self.collection_name,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from pendulum import duration

from airflow.providers.amazon.aws.hooks.eks import ClusterStates, FargateProfileStates
from airflow.providers.amazon.aws.operators.eks import (
    EksCreateClusterOperator,
    EksDeleteClusterOperator,
    EksPodOperator,
)
from airflow.providers.amazon.aws.sensors.eks import EksClusterStateSensor, EksFargateProfileStateSensor
from airflow.providers.common.compat.sdk import DAG, chain

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder
from system.amazon.aws.utils.k8s import get_describe_pod_operator

DAG_ID = "example_eks_with_fargate_in_one_step"

# Externally fetched variables
# See https://docs.aws.amazon.com/eks/latest/userguide/service_IAM_role.html
CLUSTER_ROLE_ARN_KEY = "CLUSTER_ROLE_ARN"
# See https://docs.aws.amazon.com/eks/latest/userguide/pod-execution-role.html
FARGATE_POD_ROLE_ARN_KEY = "FARGATE_POD_ROLE_ARN"
SUBNETS_KEY = "SUBNETS"

sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(CLUSTER_ROLE_ARN_KEY)
    .add_variable(FARGATE_POD_ROLE_ARN_KEY)
    .add_variable(SUBNETS_KEY, split_string=True)
    .build()
)

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    cluster_role_arn = test_context[CLUSTER_ROLE_ARN_KEY]
    fargate_pod_role_arn = test_context[FARGATE_POD_ROLE_ARN_KEY]
    subnets = test_context[SUBNETS_KEY]

    cluster_name = f"{env_id}-cluster"
    fargate_profile_name = f"{env_id}-profile"
    test_name = f"{env_id}_{DAG_ID}"

    # [START howto_operator_eks_create_cluster_with_fargate_profile]
    # Create an Amazon EKS cluster control plane and an AWS Fargate compute platform in one step.
    create_cluster_and_fargate_profile = EksCreateClusterOperator(
        task_id="create_eks_cluster_and_fargate_profile",
        cluster_name=cluster_name,
        cluster_role_arn=cluster_role_arn,
        resources_vpc_config={
            "subnetIds": subnets,
            "endpointPublicAccess": True,
            "endpointPrivateAccess": False,
        },
        compute="fargate",
        fargate_profile_name=fargate_profile_name,
        # Opting to use the same ARN for the cluster and the pod here,
        # but a different ARN could be configured and passed if desired.
        fargate_pod_execution_role_arn=fargate_pod_role_arn,
    )
    # [END howto_operator_eks_create_cluster_with_fargate_profile]

    await_create_fargate_profile = EksFargateProfileStateSensor(
        task_id="await_create_fargate_profile",
        cluster_name=cluster_name,
        fargate_profile_name=fargate_profile_name,
        target_state=FargateProfileStates.ACTIVE,
    )

    start_pod = EksPodOperator(
        task_id="run_pod",
        pod_name="run_pod",
        cluster_name=cluster_name,
        image="amazon/aws-cli:latest",
        cmds=["sh", "-c", "echo Test Airflow; date"],
        labels={"demo": "hello_world"},
        get_logs=True,
        startup_timeout_seconds=600,
        # Keep the pod alive, so we can describe it in case of trouble. It's deleted with the cluster anyway.
        on_finish_action="keep_pod",
    )

    describe_pod = get_describe_pod_operator(
        cluster_name, pod_name="{{ ti.xcom_pull(key='pod_name', task_ids='run_pod') }}"
    )
    # only describe the pod if the task above failed, to help diagnose
    describe_pod.trigger_rule = TriggerRule.ONE_FAILED

    # Wait for fargate profile to be in stable state before deletion
    await_fargate_profile_stable = EksFargateProfileStateSensor(
        task_id="await_fargate_profile_stable",
        trigger_rule=TriggerRule.ALL_DONE,
        cluster_name=cluster_name,
        fargate_profile_name=fargate_profile_name,
        target_state=FargateProfileStates.ACTIVE,
    )

    # An Amazon EKS cluster can not be deleted with attached resources such as nodegroups or Fargate profiles.
    # Setting the `force` to `True` will delete any attached resources before deleting the cluster.
    delete_cluster_and_fargate_profile = EksDeleteClusterOperator(
        task_id="delete_fargate_profile_and_cluster",
        trigger_rule=TriggerRule.ALL_DONE,
        cluster_name=cluster_name,
        force_delete_compute=True,
        retries=4,
        retry_delay=duration(seconds=30),
        retry_exponential_backoff=True,
    )

    await_delete_cluster = EksClusterStateSensor(
        task_id="await_delete_cluster",
        trigger_rule=TriggerRule.ALL_DONE,
        cluster_name=cluster_name,
        target_state=ClusterStates.NONEXISTENT,
        poke_interval=10,
    )

    chain(
        # TEST SETUP
        test_context,
        # TEST BODY
        create_cluster_and_fargate_profile,
        await_create_fargate_profile,
        start_pod,
        # TEST TEARDOWN
        describe_pod,
        await_fargate_profile_stable,
        delete_cluster_and_fargate_profile,
        await_delete_cluster,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.operators.redshift_cluster import (
    RedshiftCreateClusterOperator,
    RedshiftCreateClusterSnapshotOperator,
    RedshiftDeleteClusterOperator,
    RedshiftDeleteClusterSnapshotOperator,
    RedshiftPauseClusterOperator,
    RedshiftResumeClusterOperator,
)
from airflow.providers.amazon.aws.operators.redshift_data import RedshiftDataOperator
from airflow.providers.amazon.aws.sensors.redshift_cluster import RedshiftClusterSensor
from airflow.providers.common.compat.sdk import DAG, chain

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_redshift"

# Externally fetched variables:
SECURITY_GROUP_KEY = "SECURITY_GROUP"
CLUSTER_SUBNET_GROUP_KEY = "CLUSTER_SUBNET_GROUP"

sys_test_context_task = (
    SystemTestContextBuilder().add_variable(SECURITY_GROUP_KEY).add_variable(CLUSTER_SUBNET_GROUP_KEY).build()
)

DB_LOGIN = "adminuser"
DB_PASS = "MyAmazonPassword1"
DB_NAME = "dev"
POLL_INTERVAL = 10

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    security_group_id = test_context[SECURITY_GROUP_KEY]
    cluster_subnet_group_name = test_context[CLUSTER_SUBNET_GROUP_KEY]
    redshift_cluster_identifier = f"{env_id}-redshift-cluster"
    redshift_cluster_snapshot_identifier = f"{env_id}-snapshot"
    sg_name = f"{env_id}-sg"

    # [START howto_operator_redshift_cluster]
    create_cluster = RedshiftCreateClusterOperator(
        task_id="create_cluster",
        cluster_identifier=redshift_cluster_identifier,
        vpc_security_group_ids=[security_group_id],
        cluster_subnet_group_name=cluster_subnet_group_name,
        publicly_accessible=False,
        cluster_type="single-node",
        node_type="ra3.large",
        master_username=DB_LOGIN,
        master_user_password=DB_PASS,
    )
    # [END howto_operator_redshift_cluster]

    # [START howto_sensor_redshift_cluster]
    wait_cluster_available = RedshiftClusterSensor(
        task_id="wait_cluster_available",
        cluster_identifier=redshift_cluster_identifier,
        target_status="available",
        poke_interval=15,
        timeout=60 * 30,
    )
    # [END howto_sensor_redshift_cluster]

    # [START howto_operator_redshift_create_cluster_snapshot]
    create_cluster_snapshot = RedshiftCreateClusterSnapshotOperator(
        task_id="create_cluster_snapshot",
        cluster_identifier=redshift_cluster_identifier,
        snapshot_identifier=redshift_cluster_snapshot_identifier,
        poll_interval=30,
        max_attempt=100,
        retention_period=1,
        wait_for_completion=True,
    )
    # [END howto_operator_redshift_create_cluster_snapshot]

    wait_cluster_available_before_pause = RedshiftClusterSensor(
        task_id="wait_cluster_available_before_pause",
        cluster_identifier=redshift_cluster_identifier,
        target_status="available",
        poke_interval=15,
        timeout=60 * 30,
    )

    # [START howto_operator_redshift_pause_cluster]
    pause_cluster = RedshiftPauseClusterOperator(
        task_id="pause_cluster",
        cluster_identifier=redshift_cluster_identifier,
    )
    # [END howto_operator_redshift_pause_cluster]

    wait_cluster_paused = RedshiftClusterSensor(
        task_id="wait_cluster_paused",
        cluster_identifier=redshift_cluster_identifier,
        target_status="paused",
        poke_interval=30,
        timeout=60 * 30,
    )

    # [START howto_operator_redshift_resume_cluster]
    resume_cluster = RedshiftResumeClusterOperator(
        task_id="resume_cluster",
        cluster_identifier=redshift_cluster_identifier,
    )
    # [END howto_operator_redshift_resume_cluster]

    wait_cluster_available_after_resume = RedshiftClusterSensor(
        task_id="wait_cluster_available_after_resume",
        cluster_identifier=redshift_cluster_identifier,
        target_status="available",
        poke_interval=30,
        timeout=60 * 30,
    )

    # [START howto_operator_redshift_data]
    create_table_redshift_data = RedshiftDataOperator(
        task_id="create_table_redshift_data",
        cluster_identifier=redshift_cluster_identifier,
        database=DB_NAME,
        db_user=DB_LOGIN,
        sql=[
            """
            CREATE TABLE IF NOT EXISTS fruit (
            fruit_id INTEGER,
            name VARCHAR NOT NULL,
            color VARCHAR NOT NULL
            );
        """
        ],
        poll_interval=POLL_INTERVAL,
        wait_for_completion=True,
    )
    # [END howto_operator_redshift_data]

    insert_data = RedshiftDataOperator(
        task_id="insert_data",
        cluster_identifier=redshift_cluster_identifier,
        database=DB_NAME,
        db_user=DB_LOGIN,
        sql=[
            "INSERT INTO fruit VALUES ( 1, 'Banana', 'Yellow');",
            "INSERT INTO fruit VALUES ( 2, 'Apple', 'Red');",
            "INSERT INTO fruit VALUES ( 3, 'Lemon', 'Yellow');",
            "INSERT INTO fruit VALUES ( 4, 'Grape', 'Purple');",
            "INSERT INTO fruit VALUES ( 5, 'Pear', 'Green');",
            "INSERT INTO fruit VALUES ( 6, 'Strawberry', 'Red');",
        ],
        poll_interval=POLL_INTERVAL,
        wait_for_completion=True,
    )

    # [START howto_operator_redshift_data_session_reuse]
    create_tmp_table_data_api = RedshiftDataOperator(
        task_id="create_tmp_table_data_api",
        cluster_identifier=redshift_cluster_identifier,
        database=DB_NAME,
        db_user=DB_LOGIN,
        sql=[
            """
            CREATE TEMPORARY TABLE tmp_people (
            id INTEGER,
            first_name VARCHAR(100),
            age INTEGER
            );
        """
        ],
        poll_interval=POLL_INTERVAL,
        wait_for_completion=True,
        session_keep_alive_seconds=600,
    )

    insert_data_reuse_session = RedshiftDataOperator(
        task_id="insert_data_reuse_session",
        sql=[
            "INSERT INTO tmp_people VALUES ( 1, 'Bob', 30);",
            "INSERT INTO tmp_people VALUES ( 2, 'Alice', 35);",
            "INSERT INTO tmp_people VALUES ( 3, 'Charlie', 40);",
        ],
        poll_interval=POLL_INTERVAL,
        wait_for_completion=True,
        session_id="{{ task_instance.xcom_pull(task_ids='create_tmp_table_data_api', key='session_id') }}",
    )
    # [END howto_operator_redshift_data_session_reuse]

    # [START howto_operator_redshift_delete_cluster]
    delete_cluster = RedshiftDeleteClusterOperator(
        task_id="delete_cluster",
        cluster_identifier=redshift_cluster_identifier,
    )
    # [END howto_operator_redshift_delete_cluster]
    delete_cluster.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_operator_redshift_delete_cluster_snapshot]
    delete_cluster_snapshot = RedshiftDeleteClusterSnapshotOperator(
        task_id="delete_cluster_snapshot",
        cluster_identifier=redshift_cluster_identifier,
        snapshot_identifier=redshift_cluster_snapshot_identifier,
    )
    # [END howto_operator_redshift_delete_cluster_snapshot]

    chain(
        # TEST SETUP
        test_context,
        # TEST BODY
        create_cluster,
        wait_cluster_available,
        create_cluster_snapshot,
        wait_cluster_available_before_pause,
        pause_cluster,
        wait_cluster_paused,
        resume_cluster,
        wait_cluster_available_after_resume,
        create_table_redshift_data,
        insert_data,
        delete_cluster_snapshot,
        delete_cluster,
    )

    # Test session reuse in parallel
    chain(
        wait_cluster_available_after_resume,
        create_tmp_table_data_api,
        insert_data_reuse_session,
        delete_cluster_snapshot,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from io import RawIOBase
from unittest import mock

from moto import mock_aws

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.transfers.azure_blob_to_s3 import AzureBlobStorageToS3Operator

TASK_ID = "test-gcs-list-operator"
CONTAINER_NAME = "test-container"
DELIMITER = ".csv"
PREFIX = "TEST"
S3_BUCKET = "s3://bucket/"
MOCK_FILES = ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
S3_ACL_POLICY = "private-read"


def _create_test_bucket():
    hook = S3Hook(aws_conn_id="airflow_gcs_test")
    # We're mocking all actual AWS calls and don't need a connection.
    # This avoids an Airflow warning about connection cannot be found.
    hook.get_connection = lambda _: None
    bucket = hook.get_bucket("bucket")
    bucket.create()
    return hook, bucket


@mock_aws
class TestAzureBlobToS3Operator:
    @mock.patch("airflow.providers.amazon.aws.transfers.azure_blob_to_s3.WasbHook")
    def test_operator_all_file_upload(self, mock_hook):
        """
        Destination bucket has no file (of interest) common with origin bucket i.e
        Azure - ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
        S3 - []
        """
        mock_hook.return_value.get_blobs_list_recursive.return_value = MOCK_FILES

        operator = AzureBlobStorageToS3Operator(
            task_id=TASK_ID,
            container_name=CONTAINER_NAME,
            dest_s3_key=S3_BUCKET,
            replace=False,
        )

        hook, _ = _create_test_bucket()
        uploaded_files = operator.execute(None)

        assert sorted(MOCK_FILES) == sorted(uploaded_files)
        assert sorted(MOCK_FILES) == sorted(hook.list_keys("bucket", delimiter="/"))

    @mock.patch("airflow.providers.amazon.aws.transfers.azure_blob_to_s3.WasbHook")
    def test_operator_incremental_file_upload_without_replace(self, mock_hook):
        """
        Destination bucket has subset of files common with origin bucket i.e
        Azure - ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
        S3 - ["TEST1.csv"]
        """
        mock_hook.return_value.get_blobs_list_recursive.return_value = MOCK_FILES
        get_file = mock_hook.return_value.get_file

        operator = AzureBlobStorageToS3Operator(
            task_id=TASK_ID,
            container_name=CONTAINER_NAME,
            dest_s3_key=S3_BUCKET,
            # without replace
            replace=False,
        )

        hook, bucket = _create_test_bucket()
        # uploading only first file
        bucket.put_object(Key=MOCK_FILES[0], Body=b"testing")

        uploaded_files = operator.execute(None)

        assert sorted(MOCK_FILES[1:]) == sorted(uploaded_files)
        assert sorted(MOCK_FILES) == sorted(hook.list_keys("bucket", delimiter="/"))
        assert get_file.call_count == len(MOCK_FILES[1:])

    @mock.patch("airflow.providers.amazon.aws.transfers.azure_blob_to_s3.WasbHook")
    def test_operator_incremental_file_upload_with_replace(self, mock_hook):
        """
        Destination bucket has subset of files common with origin bucket i.e
        Azure - ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
        S3 - ["TEST1.csv"]
        """
        mock_hook.return_value.get_blobs_list_recursive.return_value = MOCK_FILES
        get_file = mock_hook.return_value.get_file

        operator = AzureBlobStorageToS3Operator(
            task_id=TASK_ID,
            container_name=CONTAINER_NAME,
            dest_s3_key=S3_BUCKET,
            # with replace
            replace=True,
        )

        hook, bucket = _create_test_bucket()
        # uploading only first file
        bucket.put_object(Key=MOCK_FILES[0], Body=b"testing")

        uploaded_files = operator.execute(None)

        assert sorted(MOCK_FILES) == sorted(uploaded_files)
        assert sorted(MOCK_FILES) == sorted(hook.list_keys("bucket", delimiter="/"))
        assert get_file.call_count == len(MOCK_FILES)

    @mock.patch("airflow.providers.amazon.aws.transfers.azure_blob_to_s3.WasbHook")
    def test_operator_no_file_upload_without_replace(self, mock_hook):
        """
        Destination bucket has all the files common with origin bucket i.e
        Azure - ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
        S3 - ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
        """
        mock_hook.return_value.get_blobs_list_recursive.return_value = MOCK_FILES
        get_file = mock_hook.return_value.get_file

        operator = AzureBlobStorageToS3Operator(
            task_id=TASK_ID,
            container_name=CONTAINER_NAME,
            dest_s3_key=S3_BUCKET,
            replace=False,
        )

        hook, bucket = _create_test_bucket()
        # uploading all the files
        for mock_file in MOCK_FILES:
            bucket.put_object(Key=mock_file, Body=b"testing")

        uploaded_files = operator.execute(None)

        assert sorted([]) == sorted(uploaded_files)
        assert sorted(MOCK_FILES) == sorted(hook.list_keys("bucket", delimiter="/"))
        assert get_file.call_count == 0

    @mock.patch("airflow.providers.amazon.aws.transfers.azure_blob_to_s3.WasbHook")
    def test_operator_no_file_upload_with_replace(self, mock_hook):
        """
        Destination bucket has all the files common with origin bucket i.e
        Azure - ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
        S3 - ["TEST1.csv", "TEST2.csv", "TEST3.csv"]
        """
        mock_hook.return_value.get_blobs_list_recursive.return_value = MOCK_FILES
        get_file = mock_hook.return_value.get_file

        operator = AzureBlobStorageToS3Operator(
            task_id=TASK_ID,
            container_name=CONTAINER_NAME,
            dest_s3_key=S3_BUCKET,
            replace=True,
        )

        hook, bucket = _create_test_bucket()
        # uploading all the files
        for mock_file in MOCK_FILES:
            bucket.put_object(Key=mock_file, Body=b"testing")

        uploaded_files = operator.execute(None)

        assert sorted(MOCK_FILES) == sorted(uploaded_files)
        assert sorted(MOCK_FILES) == sorted(hook.list_keys("bucket", delimiter="/"))
        # this ensures that upload happened
        assert get_file.call_count == len(MOCK_FILES)

    @mock.patch("tempfile.NamedTemporaryFile")
    @mock.patch("airflow.providers.amazon.aws.transfers.azure_blob_to_s3.WasbHook")
    @mock.patch("airflow.providers.amazon.aws.transfers.azure_blob_to_s3.S3Hook")
    def test_operator_should_pass_dest_s3_extra_args_and_s3_acl_policy(
        self, s3_hook_mock, wasb_hook_mock, mock_tempfile
    ):
        wasb_blob_name = "test_file"
        s3_acl_policy = "test policy"
        s3_extra_args = {"ContentLanguage": "value"}

        wasb_hook_mock.return_value.get_blobs_list_recursive.return_value = [wasb_blob_name]
        wasb_hook_mock.return_value.download.return_value = RawIOBase()
        mock_tempfile.return_value.__enter__.return_value.name = "test_temp_file"

        # with current S3_BUCKET url, parse_s3_url would complain
        s3_hook_mock.parse_s3_url.return_value = ("bucket", wasb_blob_name)
        mock_load_files = s3_hook_mock.return_value.load_file

        operator = AzureBlobStorageToS3Operator(
            task_id=TASK_ID,
            container_name=CONTAINER_NAME,
            dest_s3_key=S3_BUCKET,
            replace=False,
            dest_s3_extra_args=s3_extra_args,
            s3_acl_policy=s3_acl_policy,
        )

        operator.execute(None)
        s3_hook_mock.assert_called_once_with(aws_conn_id="aws_default", extra_args=s3_extra_args, verify=None)
        mock_load_files.assert_called_once_with(
            filename="test_temp_file",
            key=f"{S3_BUCKET}{wasb_blob_name}",
            replace=False,
            acl_policy=s3_acl_policy,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
from dateutil.tz import tzlocal

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.sensors.emr import EmrStepSensor
from airflow.providers.amazon.aws.triggers.emr import EmrStepSensorTrigger
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred

DESCRIBE_JOB_STEP_RUNNING_RETURN = {
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "8dee8db2-3719-11e6-9e20-35b2f861a2a6"},
    "Step": {
        "ActionOnFailure": "CONTINUE",
        "Config": {
            "Args": ["/usr/lib/spark/bin/run-example", "SparkPi", "10"],
            "Jar": "command-runner.jar",
            "Properties": {},
        },
        "Id": "s-VK57YR1Z9Z5N",
        "Name": "calculate_pi",
        "Status": {
            "State": "RUNNING",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime(2016, 6, 20, 19, 0, 18, tzinfo=tzlocal()),
                "StartDateTime": datetime(2016, 6, 20, 19, 2, 34, tzinfo=tzlocal()),
            },
        },
    },
}

DESCRIBE_JOB_STEP_CANCELLED_RETURN = {
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "8dee8db2-3719-11e6-9e20-35b2f861a2a6"},
    "Step": {
        "ActionOnFailure": "CONTINUE",
        "Config": {
            "Args": ["/usr/lib/spark/bin/run-example", "SparkPi", "10"],
            "Jar": "command-runner.jar",
            "Properties": {},
        },
        "Id": "s-VK57YR1Z9Z5N",
        "Name": "calculate_pi",
        "Status": {
            "State": "CANCELLED",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime(2016, 6, 20, 19, 0, 18, tzinfo=tzlocal()),
                "StartDateTime": datetime(2016, 6, 20, 19, 2, 34, tzinfo=tzlocal()),
            },
        },
    },
}

DESCRIBE_JOB_STEP_FAILED_RETURN = {
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "8dee8db2-3719-11e6-9e20-35b2f861a2a6"},
    "Step": {
        "ActionOnFailure": "CONTINUE",
        "Config": {
            "Args": ["/usr/lib/spark/bin/run-example", "SparkPi", "10"],
            "Jar": "command-runner.jar",
            "Properties": {},
        },
        "Id": "s-VK57YR1Z9Z5N",
        "Name": "calculate_pi",
        "Status": {
            "State": "FAILED",
            "StateChangeReason": {},
            "FailureDetails": {
                "LogFile": "s3://fake-log-files/emr-logs/j-8989898989/steps/s-VK57YR1Z9Z5N",
                "Reason": "Unknown Error.",
            },
            "Timeline": {
                "CreationDateTime": datetime(2016, 6, 20, 19, 0, 18, tzinfo=tzlocal()),
                "StartDateTime": datetime(2016, 6, 20, 19, 2, 34, tzinfo=tzlocal()),
            },
        },
    },
}

DESCRIBE_JOB_STEP_INTERRUPTED_RETURN = {
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "8dee8db2-3719-11e6-9e20-35b2f861a2a6"},
    "Step": {
        "ActionOnFailure": "CONTINUE",
        "Config": {
            "Args": ["/usr/lib/spark/bin/run-example", "SparkPi", "10"],
            "Jar": "command-runner.jar",
            "Properties": {},
        },
        "Id": "s-VK57YR1Z9Z5N",
        "Name": "calculate_pi",
        "Status": {
            "State": "INTERRUPTED",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime(2016, 6, 20, 19, 0, 18, tzinfo=tzlocal()),
                "StartDateTime": datetime(2016, 6, 20, 19, 2, 34, tzinfo=tzlocal()),
            },
        },
    },
}

DESCRIBE_JOB_STEP_COMPLETED_RETURN = {
    "ResponseMetadata": {"HTTPStatusCode": 200, "RequestId": "8dee8db2-3719-11e6-9e20-35b2f861a2a6"},
    "Step": {
        "ActionOnFailure": "CONTINUE",
        "Config": {
            "Args": ["/usr/lib/spark/bin/run-example", "SparkPi", "10"],
            "Jar": "command-runner.jar",
            "Properties": {},
        },
        "Id": "s-VK57YR1Z9Z5N",
        "Name": "calculate_pi",
        "Status": {
            "State": "COMPLETED",
            "StateChangeReason": {},
            "Timeline": {
                "CreationDateTime": datetime(2016, 6, 20, 19, 0, 18, tzinfo=tzlocal()),
                "StartDateTime": datetime(2016, 6, 20, 19, 2, 34, tzinfo=tzlocal()),
            },
        },
    },
}


@pytest.fixture
def mocked_hook_client():
    with mock.patch("airflow.providers.amazon.aws.hooks.emr.EmrHook.conn") as m:
        yield m


class TestEmrStepSensor:
    def setup_method(self):
        self.sensor = EmrStepSensor(
            task_id="test_task",
            poke_interval=0,
            job_flow_id="j-8989898989",
            step_id="s-VK57YR1Z9Z5N",
            aws_conn_id="aws_default",
        )

    def test_step_completed(self, mocked_hook_client):
        mocked_hook_client.describe_step.side_effect = [
            DESCRIBE_JOB_STEP_RUNNING_RETURN,
            DESCRIBE_JOB_STEP_COMPLETED_RETURN,
        ]

        with patch.object(S3Hook, "parse_s3_url", return_value="valid_uri"):
            self.sensor.execute(MagicMock())

        assert mocked_hook_client.describe_step.call_count == 2
        calls = [mock.call(ClusterId="j-8989898989", StepId="s-VK57YR1Z9Z5N")] * 2
        mocked_hook_client.describe_step.assert_has_calls(calls)

    def test_step_cancelled(self, mocked_hook_client):
        mocked_hook_client.describe_step.side_effect = [
            DESCRIBE_JOB_STEP_RUNNING_RETURN,
            DESCRIBE_JOB_STEP_CANCELLED_RETURN,
        ]

        with pytest.raises(AirflowException, match="EMR job failed"):
            self.sensor.execute(MagicMock())

    def test_step_failed(self, mocked_hook_client):
        mocked_hook_client.describe_step.side_effect = [
            DESCRIBE_JOB_STEP_RUNNING_RETURN,
            DESCRIBE_JOB_STEP_FAILED_RETURN,
        ]

        with pytest.raises(AirflowException, match="EMR job failed"):
            self.sensor.execute(MagicMock())

    def test_step_interrupted(self, mocked_hook_client):
        mocked_hook_client.describe_step.side_effect = [
            DESCRIBE_JOB_STEP_RUNNING_RETURN,
            DESCRIBE_JOB_STEP_INTERRUPTED_RETURN,
        ]

        with pytest.raises(AirflowException):
            self.sensor.execute(MagicMock())

    def test_sensor_defer(self):
        """Test the execute method raise TaskDeferred if running sensor in deferrable mode"""
        sensor = EmrStepSensor(
            task_id="test_task",
            poke_interval=0,
            job_flow_id="j-8989898989",
            step_id="s-VK57YR1Z9Z5N",
            aws_conn_id="aws_default",
            deferrable=True,
        )

        with patch.object(EmrStepSensor, "poke", return_value=False):
            with pytest.raises(TaskDeferred) as exc:
                sensor.execute(context=None)
        assert isinstance(exc.value.trigger, EmrStepSensorTrigger), "Trigger is not a EmrStepSensorTrigger"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import os
import warnings
from importlib import metadata

import pytest


@pytest.fixture(scope="session")
def botocore_version():
    try:
        version = metadata.version("botocore")
    except ModuleNotFoundError:
        warnings.warn("'botocore' package not found'", UserWarning, stacklevel=2)
        return None

    try:
        return tuple(map(int, version.split(".")[:3]))
    except Exception:
        warnings.warn(f"Unable to parse botocore {version!r}", UserWarning, stacklevel=2)
        return None


@pytest.fixture(autouse=True)
def filter_botocore_warnings(botocore_version):
    """Filter known botocore future warnings."""

    with warnings.catch_warnings():
        if botocore_version and botocore_version < (1, 29):
            # By default, for some clients botocore use deprecated endpoints `{region}.{service}.{dnsSuffix}`
            # In botocore 1.29 it will be replaced by `{service}.{region}.{dnsSuffix}`
            # and the warning should gone
            # See: https://github.com/boto/botocore/issues/2705
            warnings.filterwarnings(
                "ignore",
                category=FutureWarning,
                module="botocore.client",
                message="The .* client is currently using a deprecated endpoint.*",
            )
        yield


@pytest.fixture(scope="package")
def aws_testing_env_vars(tmp_path_factory):
    """Session scoped fixture, return mock AWS specific environment variables for unit tests."""
    tmp_dir = tmp_path_factory.mktemp("aws-configs-")

    def empty_config(name: str) -> str:
        config = tmp_dir / name
        config.touch()
        return str(config)

    return {
        # Mock values for access_key, secret_key and token
        "AWS_ACCESS_KEY_ID": "airflow-testing",
        "AWS_SECRET_ACCESS_KEY": "airflow-testing",
        "AWS_SESSION_TOKEN": "airflow-testing",
        "AWS_SECURITY_TOKEN": "airflow-testing",
        # Set default region as N.Virginia (eu-west-1).
        # Otherwise some unit tests might fail if this sets to other region.
        "AWS_DEFAULT_REGION": "us-east-1",
        "AWS_REGION": "us-east-1",
        # Create empty configuration file
        "AWS_SHARED_CREDENTIALS_FILE": empty_config("aws_shared_credentials_file"),
        "AWS_CONFIG_FILE": empty_config("aws_config_file"),
        "BOTO_CONFIG": empty_config("legacy_boto2_config.cfg"),
    }


@pytest.fixture(autouse=True)
def set_default_aws_settings(aws_testing_env_vars, monkeypatch):
    """
    Change AWS configurations (env vars) before start each test.
    1. Remove all existed variables which prefixed by AWS.
        It might be some credentials, botocore configurations, etc.
    2. Use pre-defined variables for unit testing.
    """
    for env_name in os.environ:
        if env_name.startswith("AWS"):
            monkeypatch.delenv(env_name, raising=False)
    for env_name, value in aws_testing_env_vars.items():
        monkeypatch.setenv(env_name, value)


@pytest.fixture(scope="package", autouse=True)
def setup_default_aws_connections():
    with pytest.MonkeyPatch.context() as mp_ctx:
        mp_ctx.setenv("AIRFLOW_CONN_AWS_DEFAULT", '{"conn_type": "aws"}')
        mp_ctx.setenv("AIRFLOW_CONN_EMR_DEFAULT", '{"conn_type": "emr", "extra": {}}')
        yield



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.transfers.ftp_to_s3 import FTPToS3Operator
from airflow.providers.common.compat.sdk import DAG, chain

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_ftp_to_s3"

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    s3_bucket = f"{env_id}-ftp-to-s3-bucket"
    s3_key = f"{env_id}-ftp-to-s3-key"

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=s3_bucket)

    # [START howto_transfer_ftp_to_s3]
    ftp_to_s3_task = FTPToS3Operator(
        task_id="ftp_to_s3_task",
        ftp_path="/tmp/ftp_path",
        s3_bucket=s3_bucket,
        s3_key=s3_key,
        replace=True,
    )
    # [END howto_transfer_ftp_to_s3]

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=s3_bucket,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_s3_bucket,
        # TEST BODY
        ftp_to_s3_task,
        # TEST TEARDOWN
        delete_s3_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.amazon.aws.triggers.ec2 import EC2StateSensorTrigger
from airflow.triggers.base import TriggerEvent

TEST_INSTANCE_ID = "test-instance-id"
TEST_TARGET_STATE = "test_state"
TEST_CONN_ID = "test_conn_id"
TEST_REGION_NAME = "test-region"
TEST_POLL_INTERVAL = 100


class TestEC2StateSensorTrigger:
    def test_ec2_state_sensor_trigger_serialize(self):
        test_ec2_state_sensor = EC2StateSensorTrigger(
            instance_id=TEST_INSTANCE_ID,
            target_state=TEST_TARGET_STATE,
            aws_conn_id=TEST_CONN_ID,
            region_name=TEST_REGION_NAME,
            poll_interval=TEST_POLL_INTERVAL,
        )

        class_path, args = test_ec2_state_sensor.serialize()
        assert class_path == "airflow.providers.amazon.aws.triggers.ec2.EC2StateSensorTrigger"
        assert args["instance_id"] == TEST_INSTANCE_ID
        assert args["target_state"] == TEST_TARGET_STATE
        assert args["aws_conn_id"] == TEST_CONN_ID
        assert args["region_name"] == TEST_REGION_NAME
        assert args["poll_interval"] == TEST_POLL_INTERVAL

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.amazon.aws.hooks.ec2.EC2Hook.get_instance_state_async")
    @mock.patch("airflow.providers.amazon.aws.hooks.ec2.EC2Hook.get_async_conn")
    async def test_ec2_state_sensor_run(self, mock_async_conn, mock_get_instance_state_async):
        mock = AsyncMock()
        mock_async_conn.return_value.__aenter__.return_value = mock
        mock_get_instance_state_async.return_value = TEST_TARGET_STATE

        test_ec2_state_sensor = EC2StateSensorTrigger(
            instance_id=TEST_INSTANCE_ID,
            target_state=TEST_TARGET_STATE,
            aws_conn_id=TEST_CONN_ID,
            region_name=TEST_REGION_NAME,
            poll_interval=TEST_POLL_INTERVAL,
        )

        generator = test_ec2_state_sensor.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "message": "target state met"})

    @pytest.mark.asyncio
    @mock.patch("asyncio.sleep")
    @mock.patch("airflow.providers.amazon.aws.hooks.ec2.EC2Hook.get_instance_state_async")
    @mock.patch("airflow.providers.amazon.aws.hooks.ec2.EC2Hook.get_async_conn")
    async def test_ec2_state_sensor_run_multiple(
        self, mock_async_conn, mock_get_instance_state_async, mock_sleep
    ):
        mock = AsyncMock()
        mock_async_conn.return_value.__aenter__.return_value = mock
        mock_get_instance_state_async.side_effect = ["test-state", TEST_TARGET_STATE]
        mock_sleep.return_value = True

        test_ec2_state_sensor = EC2StateSensorTrigger(
            instance_id=TEST_INSTANCE_ID,
            target_state=TEST_TARGET_STATE,
            aws_conn_id=TEST_CONN_ID,
            region_name=TEST_REGION_NAME,
            poll_interval=TEST_POLL_INTERVAL,
        )

        generator = test_ec2_state_sensor.run()
        response = await generator.asend(None)

        assert mock_get_instance_state_async.call_count == 2

        assert response == TriggerEvent({"status": "success", "message": "target state met"})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
AWS ECS Executor Boto Schema.

Schemas for easily and consistently parsing boto responses.
"""

from __future__ import annotations

from marshmallow import EXCLUDE, Schema, fields, post_load


class BotoContainerSchema(Schema):
    """
    Botocore Serialization Object for ECS ``Container`` shape.

    Note that there are many more parameters, but the executor only needs the members listed below.
    """

    exit_code = fields.Integer(data_key="exitCode")
    last_status = fields.String(data_key="lastStatus")
    name = fields.String(required=True)
    reason = fields.String()
    container_arn = fields.String(data_key="containerArn")

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE


class BotoTaskSchema(Schema):
    """
    Botocore Serialization Object for ECS ``Task`` shape.

    Note that there are many more parameters, but the executor only needs the members listed below.
    """

    task_arn = fields.String(data_key="taskArn", required=True)
    last_status = fields.String(data_key="lastStatus", required=True)
    desired_status = fields.String(data_key="desiredStatus", required=True)
    containers = fields.List(fields.Nested(BotoContainerSchema), required=True)
    started_at = fields.Raw(data_key="startedAt")
    stopped_reason = fields.String(data_key="stoppedReason")

    @post_load
    def make_task(self, data, **kwargs):
        """Overwrite marshmallow load() to return an EcsExecutorTask instance instead of a dictionary."""
        # Imported here to avoid circular import.
        from airflow.providers.amazon.aws.executors.ecs.utils import EcsExecutorTask

        return EcsExecutorTask(**data)

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE


class BotoFailureSchema(Schema):
    """Botocore Serialization Object for ECS ``Failure`` Shape."""

    arn = fields.String()
    reason = fields.String()

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE


class BotoRunTaskSchema(Schema):
    """Botocore Serialization Object for ECS ``RunTask`` Operation output."""

    tasks = fields.List(fields.Nested(BotoTaskSchema), required=True)
    failures = fields.List(fields.Nested(BotoFailureSchema), required=True)

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE


class BotoDescribeTasksSchema(Schema):
    """Botocore Serialization Object for ECS ``DescribeTask`` Operation output."""

    tasks = fields.List(fields.Nested(BotoTaskSchema), required=True)
    failures = fields.List(fields.Nested(BotoFailureSchema), required=True)

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.operators.step_function import (
    StepFunctionGetExecutionOutputOperator,
    StepFunctionStartExecutionOperator,
)
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

EXECUTION_ARN = (
    "arn:aws:states:us-east-1:123456789012:execution:"
    "pseudo-state-machine:020f5b16-b1a1-4149-946f-92dd32d97934"
)
AWS_CONN_ID = "aws_non_default"
REGION_NAME = "us-west-2"
STATE_MACHINE_ARN = "arn:aws:states:us-east-1:000000000000:stateMachine:pseudo-state-machine"
NAME = "NAME"
INPUT = "{}"


@pytest.fixture
def mocked_context():
    return mock.MagicMock(name="FakeContext")


class TestStepFunctionGetExecutionOutputOperator:
    TASK_ID = "step_function_get_execution_output"

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self):
        with mock.patch(
            "airflow.providers.amazon.aws.links.step_function.StateMachineExecutionsDetailsLink.persist"
        ) as executions_details_link:
            self.mocked_executions_details_link = executions_details_link
            yield

    def test_init(self):
        op = StepFunctionGetExecutionOutputOperator(
            task_id=self.TASK_ID,
            execution_arn=EXECUTION_ARN,
            aws_conn_id=AWS_CONN_ID,
            region_name=REGION_NAME,
            verify="/spam/egg.pem",
            botocore_config={"read_timeout": 42},
        )
        assert op.execution_arn == EXECUTION_ARN
        assert op.hook.aws_conn_id == AWS_CONN_ID
        assert op.hook._region_name == REGION_NAME
        assert op.hook._verify == "/spam/egg.pem"
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = StepFunctionGetExecutionOutputOperator(task_id=self.TASK_ID, execution_arn=EXECUTION_ARN)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(StepFunctionGetExecutionOutputOperator, "hook")
    @pytest.mark.parametrize(
        ("response", "expected_output"),
        [
            pytest.param({"output": '{"foo": "bar"}'}, {"foo": "bar"}, id="output"),
            pytest.param({"error": '{"spam": "egg"}'}, {"spam": "egg"}, id="error"),
            pytest.param({"other": '{"baz": "qux"}'}, None, id="other"),
        ],
    )
    def test_execute(self, mocked_hook, mocked_context, response, expected_output):
        mocked_hook.describe_execution.return_value = response
        op = StepFunctionGetExecutionOutputOperator(
            task_id=self.TASK_ID,
            execution_arn=EXECUTION_ARN,
            aws_conn_id=None,
        )
        assert op.execute(mocked_context) == expected_output
        mocked_hook.describe_execution.assert_called_once_with(EXECUTION_ARN)
        self.mocked_executions_details_link.assert_called_once_with(
            aws_partition=mock.ANY,
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            execution_arn=EXECUTION_ARN,
        )

    def test_template_fields(self):
        operator = StepFunctionGetExecutionOutputOperator(
            task_id=self.TASK_ID,
            execution_arn=EXECUTION_ARN,
            aws_conn_id=None,
        )
        validate_template_fields(operator)


class TestStepFunctionStartExecutionOperator:
    TASK_ID = "step_function_start_execution_task"

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self):
        with (
            mock.patch(
                "airflow.providers.amazon.aws.links.step_function.StateMachineExecutionsDetailsLink.persist"
            ) as executions_details_link,
            mock.patch(
                "airflow.providers.amazon.aws.links.step_function.StateMachineDetailsLink.persist"
            ) as details_link,
        ):
            self.mocked_executions_details_link = executions_details_link
            self.mocked_details_link = details_link
            yield

    def test_init(self):
        op = StepFunctionStartExecutionOperator(
            task_id=self.TASK_ID,
            state_machine_arn=STATE_MACHINE_ARN,
            name=NAME,
            state_machine_input=INPUT,
            aws_conn_id=AWS_CONN_ID,
            region_name=REGION_NAME,
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.state_machine_arn == STATE_MACHINE_ARN
        assert op.state_machine_arn == STATE_MACHINE_ARN
        assert op.name == NAME
        assert op.input == INPUT
        assert op.hook.aws_conn_id == AWS_CONN_ID
        assert op.hook._region_name == REGION_NAME
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

        op = StepFunctionStartExecutionOperator(task_id=self.TASK_ID, state_machine_arn=STATE_MACHINE_ARN)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

    @mock.patch.object(StepFunctionStartExecutionOperator, "hook")
    def test_execute(self, mocked_hook, mocked_context):
        hook_response = (
            "arn:aws:states:us-east-1:123456789012:execution:"
            "pseudo-state-machine:020f5b16-b1a1-4149-946f-92dd32d97934"
        )
        mocked_hook.start_execution.return_value = hook_response
        op = StepFunctionStartExecutionOperator(
            task_id=self.TASK_ID,
            state_machine_arn=STATE_MACHINE_ARN,
            name=NAME,
            state_machine_input=INPUT,
            aws_conn_id=None,
        )
        assert op.execute(mocked_context) == hook_response
        mocked_hook.start_execution.assert_called_once_with(STATE_MACHINE_ARN, NAME, INPUT, False)
        self.mocked_details_link.assert_called_once_with(
            aws_partition=mock.ANY,
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            state_machine_arn=STATE_MACHINE_ARN,
        )
        self.mocked_executions_details_link.assert_called_once_with(
            aws_partition=mock.ANY,
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            execution_arn=EXECUTION_ARN,
        )

    @mock.patch.object(StepFunctionStartExecutionOperator, "hook")
    def test_step_function_start_execution_deferrable(self, mocked_hook):
        mocked_hook.start_execution.return_value = "test-execution-arn"
        operator = StepFunctionStartExecutionOperator(
            task_id=self.TASK_ID,
            state_machine_arn=STATE_MACHINE_ARN,
            name=NAME,
            state_machine_input=INPUT,
            aws_conn_id=AWS_CONN_ID,
            region_name=REGION_NAME,
            deferrable=True,
        )
        with pytest.raises(TaskDeferred):
            operator.execute(None)
        mocked_hook.start_execution.assert_called_once_with(STATE_MACHINE_ARN, NAME, INPUT, False)

    @mock.patch.object(StepFunctionStartExecutionOperator, "hook")
    @pytest.mark.parametrize("execution_arn", [pytest.param(None, id="none"), pytest.param("", id="empty")])
    def test_step_function_no_execution_arn_returns(self, mocked_hook, execution_arn):
        mocked_hook.start_execution.return_value = execution_arn
        op = StepFunctionStartExecutionOperator(
            task_id=self.TASK_ID, state_machine_arn=STATE_MACHINE_ARN, aws_conn_id=None
        )
        with pytest.raises(AirflowException, match="Failed to start State Machine execution"):
            op.execute({})

    @mock.patch.object(StepFunctionStartExecutionOperator, "hook")
    def test_start_redrive_execution(self, mocked_hook, mocked_context):
        hook_response = (
            "arn:aws:states:us-east-1:123456789012:execution:"
            "pseudo-state-machine:020f5b16-b1a1-4149-946f-92dd32d97934"
        )
        mocked_hook.start_execution.return_value = hook_response
        op = StepFunctionStartExecutionOperator(
            task_id=self.TASK_ID,
            state_machine_arn=STATE_MACHINE_ARN,
            name=NAME,
            is_redrive_execution=True,
            state_machine_input=None,
            aws_conn_id=None,
        )
        assert op.execute(mocked_context) == hook_response
        mocked_hook.start_execution.assert_called_once_with(STATE_MACHINE_ARN, NAME, None, True)
        self.mocked_details_link.assert_called_once_with(
            aws_partition=mock.ANY,
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            state_machine_arn=STATE_MACHINE_ARN,
        )
        self.mocked_executions_details_link.assert_called_once_with(
            aws_partition=mock.ANY,
            context=mock.ANY,
            operator=mock.ANY,
            region_name=mock.ANY,
            execution_arn=EXECUTION_ARN,
        )

    def test_template_fields(self):
        operator = StepFunctionStartExecutionOperator(
            task_id=self.TASK_ID,
            state_machine_arn=STATE_MACHINE_ARN,
            name=NAME,
            is_redrive_execution=True,
            state_machine_input=None,
            aws_conn_id=None,
        )
        validate_template_fields(operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.executors.utils.base_config_keys import BaseConfigKeys
from airflow.utils.state import State

if TYPE_CHECKING:
    from airflow.models.taskinstance import TaskInstanceKey

CommandType = list[str]
ExecutorConfigType = dict[str, Any]

CONFIG_GROUP_NAME = "aws_batch_executor"

CONFIG_DEFAULTS = {
    "conn_id": "aws_default",
    "max_submit_job_attempts": "3",
    "check_health_on_startup": "True",
}


@dataclass
class BatchQueuedJob:
    """Represents a Batch job that is queued. The job will be run in the next heartbeat."""

    key: TaskInstanceKey
    command: CommandType
    queue: str
    executor_config: ExecutorConfigType
    attempt_number: int
    next_attempt_time: datetime.datetime


@dataclass
class BatchJobInfo:
    """Contains information about a currently running Batch job."""

    cmd: CommandType
    queue: str
    config: ExecutorConfigType


class BatchJob:
    """Data Transfer Object for an AWS Batch Job."""

    STATE_MAPPINGS = {
        "SUBMITTED": State.QUEUED,
        "PENDING": State.QUEUED,
        "RUNNABLE": State.QUEUED,
        "STARTING": State.QUEUED,
        "RUNNING": State.RUNNING,
        "SUCCEEDED": State.SUCCESS,
        "FAILED": State.FAILED,
    }

    def __init__(self, job_id: str, status: str, status_reason: str | None = None):
        self.job_id = job_id
        self.status = status
        self.status_reason = status_reason

    def get_job_state(self) -> str:
        """Return the state of the job."""
        return self.STATE_MAPPINGS.get(self.status, State.QUEUED)

    def __repr__(self):
        """Return a visual representation of the Job status."""
        return f"({self.job_id} -> {self.status}, {self.get_job_state()})"


class BatchJobCollection:
    """A collection to manage running Batch Jobs."""

    def __init__(self):
        self.key_to_id: dict[TaskInstanceKey, str] = {}
        self.id_to_key: dict[str, TaskInstanceKey] = {}
        self.id_to_failure_counts: dict[str, int] = defaultdict(int)
        self.id_to_job_info: dict[str, BatchJobInfo] = {}

    def add_job(
        self,
        job_id: str,
        airflow_task_key: TaskInstanceKey,
        airflow_cmd: CommandType,
        queue: str,
        exec_config: ExecutorConfigType,
        attempt_number: int,
    ):
        """Add a job to the collection."""
        self.key_to_id[airflow_task_key] = job_id
        self.id_to_key[job_id] = airflow_task_key
        self.id_to_failure_counts[job_id] = attempt_number
        self.id_to_job_info[job_id] = BatchJobInfo(cmd=airflow_cmd, queue=queue, config=exec_config)

    def pop_by_id(self, job_id: str) -> TaskInstanceKey:
        """Delete job from collection based off of Batch Job ID."""
        task_key = self.id_to_key[job_id]
        del self.key_to_id[task_key]
        del self.id_to_key[job_id]
        del self.id_to_failure_counts[job_id]
        return task_key

    def failure_count_by_id(self, job_id: str) -> int:
        """Get the number of times a job has failed given a Batch Job Id."""
        return self.id_to_failure_counts[job_id]

    def increment_failure_count(self, job_id: str):
        """Increment the failure counter given a Batch Job Id."""
        self.id_to_failure_counts[job_id] += 1

    def get_all_jobs(self) -> list[str]:
        """Get all AWS ARNs in collection."""
        return list(self.id_to_key.keys())

    def __len__(self):
        """Return the number of jobs in collection."""
        return len(self.key_to_id)


class BatchSubmitJobKwargsConfigKeys(BaseConfigKeys):
    """Keys loaded into the config which are valid Batch submit_job kwargs."""

    JOB_NAME = "job_name"
    JOB_QUEUE = "job_queue"
    JOB_DEFINITION = "job_definition"
    EKS_PROPERTIES_OVERRIDE = "eks_properties_override"
    NODE_OVERRIDE = "node_override"


class AllBatchConfigKeys(BatchSubmitJobKwargsConfigKeys):
    """All keys loaded into the config which are related to the Batch Executor."""

    MAX_SUBMIT_JOB_ATTEMPTS = "max_submit_job_attempts"
    AWS_CONN_ID = "conn_id"
    SUBMIT_JOB_KWARGS = "submit_job_kwargs"
    REGION_NAME = "region_name"
    CHECK_HEALTH_ON_STARTUP = "check_health_on_startup"


class BatchExecutorException(Exception):
    """Thrown when something unexpected has occurred within the AWS Batch ecosystem."""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from unittest.mock import MagicMock, call

import boto3
import pytest
from moto import mock_aws

import airflow.version
from airflow.models import Connection
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.common.compat.sdk import AirflowException

from tests_common.test_utils.config import conf_vars

AWS_CONN_ID_WITH_REGION = "s3_dags_connection"
AWS_CONN_ID_REGION = "eu-central-1"
AWS_CONN_ID_DEFAULT = "aws_default"
S3_BUCKET_NAME = "my-airflow-dags-bucket"
S3_BUCKET_PREFIX = "project1/dags"

if airflow.version.version.strip().startswith("3"):
    from airflow.providers.amazon.aws.bundles.s3 import S3DagBundle


@pytest.fixture
def mocked_s3_resource():
    with mock_aws():
        yield boto3.resource("s3")


@pytest.fixture
def s3_client():
    with mock_aws():
        yield boto3.client("s3")


@pytest.fixture
def s3_bucket(mocked_s3_resource, s3_client):
    bucket = mocked_s3_resource.create_bucket(Bucket=S3_BUCKET_NAME)

    s3_client.put_object(Bucket=bucket.name, Key=S3_BUCKET_PREFIX + "/dag_01.py", Body=b"test data")
    s3_client.put_object(Bucket=bucket.name, Key=S3_BUCKET_PREFIX + "/dag_02.py", Body=b"test data")
    s3_client.put_object(
        Bucket=bucket.name, Key=S3_BUCKET_PREFIX + "/subproject1/dag_a.py", Body=b"test data"
    )
    s3_client.put_object(
        Bucket=bucket.name, Key=S3_BUCKET_PREFIX + "/subproject1/dag_b.py", Body=b"test data"
    )

    return bucket


@pytest.fixture(autouse=True)
def bundle_temp_dir(tmp_path):
    with conf_vars({("dag_processor", "dag_bundle_storage_path"): str(tmp_path)}):
        yield tmp_path


@pytest.mark.skipif(not airflow.version.version.strip().startswith("3"), reason="Airflow >=3.0.0 test")
class TestS3DagBundle:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            Connection(
                conn_id=AWS_CONN_ID_DEFAULT,
                conn_type="aws",
                extra={
                    "config_kwargs": {"s3": {"bucket_name": S3_BUCKET_NAME}},
                },
            )
        )
        create_connection_without_db(
            Connection(
                conn_id=AWS_CONN_ID_WITH_REGION,
                conn_type="aws",
                extra={
                    "config_kwargs": {"s3": {"bucket_name": S3_BUCKET_NAME}},
                    "region_name": AWS_CONN_ID_REGION,
                },
            )
        )

    def test_view_url_generates_presigned_url(self):
        bundle = S3DagBundle(
            name="test", aws_conn_id=AWS_CONN_ID_DEFAULT, prefix="project1/dags", bucket_name=S3_BUCKET_NAME
        )

        url: str = bundle.view_url("test_version")
        assert url.startswith("https://my-airflow-dags-bucket.s3.amazonaws.com/project1/dags")

    def test_view_url_template_generates_presigned_url(self):
        bundle = S3DagBundle(
            name="test", aws_conn_id=AWS_CONN_ID_DEFAULT, prefix="project1/dags", bucket_name=S3_BUCKET_NAME
        )
        url: str = bundle.view_url_template()
        assert url.startswith("https://my-airflow-dags-bucket.s3.amazonaws.com/project1/dags")

    def test_supports_versioning(self):
        bundle = S3DagBundle(
            name="test", aws_conn_id=AWS_CONN_ID_DEFAULT, prefix="project1/dags", bucket_name=S3_BUCKET_NAME
        )
        assert S3DagBundle.supports_versioning is False

        # set version, it's not supported
        bundle.version = "test_version"

        with pytest.raises(AirflowException, match="Refreshing a specific version is not supported"):
            bundle.refresh()
        with pytest.raises(AirflowException, match="S3 url with version is not supported"):
            bundle.view_url("test_version")

    def test_correct_bundle_path_used(self):
        bundle = S3DagBundle(
            name="test", aws_conn_id=AWS_CONN_ID_DEFAULT, prefix="project1_dags", bucket_name="airflow_dags"
        )
        assert str(bundle.base_dir) == str(bundle.s3_dags_dir)

    def test_s3_bucket_and_prefix_validated(self, s3_bucket):
        hook = S3Hook(aws_conn_id=AWS_CONN_ID_DEFAULT)
        assert hook.check_for_bucket(s3_bucket.name) is True

        bundle = S3DagBundle(
            name="test",
            aws_conn_id=AWS_CONN_ID_WITH_REGION,
            prefix="project1_dags",
            bucket_name="non-existing-bucket",
        )
        with pytest.raises(AirflowException, match="S3 bucket.*non-existing-bucket.*does not exist.*"):
            bundle.initialize()

        bundle = S3DagBundle(
            name="test",
            aws_conn_id=AWS_CONN_ID_WITH_REGION,
            prefix="non-existing-prefix",
            bucket_name=S3_BUCKET_NAME,
        )
        with pytest.raises(AirflowException, match="S3 prefix.*non-existing-prefix.*does not exist.*"):
            bundle.initialize()

        bundle = S3DagBundle(
            name="test",
            aws_conn_id=AWS_CONN_ID_WITH_REGION,
            prefix=S3_BUCKET_PREFIX,
            bucket_name=S3_BUCKET_NAME,
        )
        # initialize succeeds, with correct prefix and bucket
        bundle.initialize()
        assert bundle.s3_hook.region_name == AWS_CONN_ID_REGION

        bundle = S3DagBundle(
            name="test",
            aws_conn_id=AWS_CONN_ID_WITH_REGION,
            prefix="",
            bucket_name=S3_BUCKET_NAME,
        )
        # initialize succeeds, with empty prefix
        bundle.initialize()
        assert bundle.s3_hook.region_name == AWS_CONN_ID_REGION

    def _upload_fixtures(self, bucket: str, fixtures_dir: str) -> None:
        client = boto3.client("s3")
        fixtures_paths = [
            os.path.join(path, filename) for path, _, files in os.walk(fixtures_dir) for filename in files
        ]
        for path in fixtures_paths:
            key = os.path.relpath(path, fixtures_dir)
            client.upload_file(Filename=path, Bucket=bucket, Key=key)

    def test_refresh(self, s3_bucket, s3_client):
        bundle = S3DagBundle(
            name="test",
            aws_conn_id=AWS_CONN_ID_WITH_REGION,
            prefix=S3_BUCKET_PREFIX,
            bucket_name=S3_BUCKET_NAME,
        )
        bundle._log.debug = MagicMock()
        # Create a pytest Call object to compare against the call_args_list of the _log.debug mock
        download_log_call = call(
            "Downloading DAGs from s3://%s/%s to %s", S3_BUCKET_NAME, S3_BUCKET_PREFIX, bundle.s3_dags_dir
        )
        bundle.initialize()
        assert bundle._log.debug.call_count == 1
        assert bundle._log.debug.call_args_list == [download_log_call]
        bundle.refresh()
        assert bundle._log.debug.call_count == 2
        assert bundle._log.debug.call_args_list == [download_log_call, download_log_call]
        bundle.refresh()
        assert bundle._log.debug.call_count == 3
        assert bundle._log.debug.call_args_list == [download_log_call, download_log_call, download_log_call]

    def test_refresh_without_prefix(self, s3_bucket, s3_client):
        bundle = S3DagBundle(
            name="test",
            aws_conn_id=AWS_CONN_ID_WITH_REGION,
            bucket_name=S3_BUCKET_NAME,
        )
        bundle._log.debug = MagicMock()
        download_log_call = call(
            "Downloading DAGs from s3://%s/%s to %s", S3_BUCKET_NAME, "", bundle.s3_dags_dir
        )
        assert bundle.prefix == ""
        bundle.initialize()
        bundle.refresh()
        assert bundle._log.debug.call_count == 2
        assert bundle._log.debug.call_args_list == [download_log_call, download_log_call]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.providers.amazon.aws.sensors.dynamodb import DynamoDBValueSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

# TODO: FIXME The argument types here seems somewhat tricky to fix
# mypy: disable-error-code="arg-type"

DAG_ID = "example_dynamodb"
sys_test_context_task = SystemTestContextBuilder().build()

PK_ATTRIBUTE_NAME = "PK"
SK_ATTRIBUTE_NAME = "SK"
TABLE_ATTRIBUTES = [
    {"AttributeName": PK_ATTRIBUTE_NAME, "AttributeType": "S"},
    {"AttributeName": SK_ATTRIBUTE_NAME, "AttributeType": "S"},
]
TABLE_KEY_SCHEMA = [
    {"AttributeName": "PK", "KeyType": "HASH"},
    {"AttributeName": "SK", "KeyType": "RANGE"},
]
TABLE_THROUGHPUT = {"ReadCapacityUnits": 10, "WriteCapacityUnits": 10}


@task
def create_table(table_name: str):
    ddb = boto3.resource("dynamodb")
    table = ddb.create_table(
        AttributeDefinitions=TABLE_ATTRIBUTES,
        TableName=table_name,
        KeySchema=TABLE_KEY_SCHEMA,
        ProvisionedThroughput=TABLE_THROUGHPUT,
    )
    boto3.client("dynamodb").get_waiter("table_exists").wait(TableName=table_name)
    table.put_item(Item={"PK": "Test", "SK": "2022-07-12T11:11:25-0400", "Value": "Testing"})


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_table(table_name: str):
    client = boto3.client("dynamodb")
    client.delete_table(TableName=table_name)
    client.get_waiter("table_not_exists").wait(TableName=table_name)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    table_name = f"{env_id}-dynamodb-table"
    create_table = create_table(table_name=table_name)
    delete_table = delete_table(table_name)

    # [START howto_sensor_dynamodb_value]
    dynamodb_sensor = DynamoDBValueSensor(
        task_id="waiting_for_dynamodb_item_value",
        table_name=table_name,
        partition_key_name=PK_ATTRIBUTE_NAME,
        partition_key_value="Test",
        sort_key_name=SK_ATTRIBUTE_NAME,
        sort_key_value="2022-07-12T11:11:25-0400",
        attribute_name="Value",
        attribute_value="Testing",
    )
    # [END howto_sensor_dynamodb_value]

    # [START howto_sensor_dynamodb_any_value]
    dynamodb_sensor_any_value = DynamoDBValueSensor(
        task_id="waiting_for_dynamodb_item_any_value",
        table_name=table_name,
        partition_key_name=PK_ATTRIBUTE_NAME,
        partition_key_value="Test",
        sort_key_name=SK_ATTRIBUTE_NAME,
        sort_key_value="2022-07-12T11:11:25-0400",
        attribute_name="Value",
        attribute_value=["Foo", "Testing", "Bar"],
    )
    # [END howto_sensor_dynamodb_any_value]

    chain(
        # TEST SETUP
        test_context,
        create_table,
        # TEST BODY
        dynamodb_sensor,
        dynamodb_sensor_any_value,
        # TEST TEARDOWN
        delete_table,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.hooks.sagemaker import SageMakerHook
from airflow.providers.amazon.aws.sensors.sagemaker import SageMakerAutoMLSensor
from airflow.providers.common.compat.sdk import AirflowException


class TestSageMakerAutoMLSensor:
    @staticmethod
    def get_response_with_state(state: str):
        states = {"Completed", "InProgress", "Failed", "Stopped", "Stopping"}
        assert state in states
        return {
            "AutoMLJobStatus": state,
            "AutoMLJobSecondaryStatus": "Starting",
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
            },
        }

    @mock.patch.object(SageMakerHook, "_describe_auto_ml_job")
    def test_sensor_with_failure(self, mock_describe):
        mock_describe.return_value = self.get_response_with_state("Failed")
        sensor = SageMakerAutoMLSensor(job_name="job_job", task_id="test_task")

        with pytest.raises(AirflowException):
            sensor.execute(None)

        mock_describe.assert_called_once_with("job_job")

    @mock.patch.object(SageMakerHook, "_describe_auto_ml_job")
    def test_sensor(self, mock_describe):
        mock_describe.side_effect = [
            self.get_response_with_state("InProgress"),
            self.get_response_with_state("Stopping"),
            self.get_response_with_state("Stopped"),
        ]
        sensor = SageMakerAutoMLSensor(job_name="job_job", task_id="test_task", poke_interval=0)

        sensor.execute(None)

        assert mock_describe.call_count == 3



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import json
from datetime import datetime

import boto3

from airflow.providers.amazon.aws.operators.quicksight import QuickSightCreateIngestionOperator
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.quicksight import QuickSightSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

"""
Prerequisites:
1. The account which runs this test must manually be activated in Quicksight here:
https://quicksight.aws.amazon.com/sn/console/signup?#
2. The activation process creates an IAM Role called `aws-quicksight-service-role-v0`.
 You have to add a policy named 'AWSQuickSightS3Policy' with the S3 access permissions.
 The policy name is enforced, and the permissions json can be copied from `AmazonS3FullAccess`.
3. Navigate to your QuickSight account, select your user account and click "Manage QuickSight".
 Then select 'Security & Permissions' and under "IAM role in use" specify the
 `aws-quicksight-service-role-v0` role.

NOTES:  If Create Ingestion fails for any reason, that ingestion name will remain in use and
 future runs will stall with the sensor returning a status of QUEUED "forever".  If you run
 into this behavior, changing the template for the ingestion name or the ENV_ID and re-running
 the test should resolve the issue.
"""

DAG_ID = "example_quicksight"

sys_test_context_task = SystemTestContextBuilder().build()

SAMPLE_DATA_COLUMNS = ["Project", "Year"]
SAMPLE_DATA = """'Airflow','2015'
    'OpenOffice','2012'
    'Subversion','2000'
    'NiFi','2006'
"""


@task
def get_aws_account_id() -> int:
    return boto3.client("sts").get_caller_identity()["Account"]


@task
def create_quicksight_data_source(
    aws_account_id: str, datasource_name: str, bucket: str, manifest_key: str
) -> str:
    response = boto3.client("quicksight").create_data_source(
        AwsAccountId=aws_account_id,
        DataSourceId=datasource_name,
        Name=datasource_name,
        Type="S3",
        DataSourceParameters={
            "S3Parameters": {"ManifestFileLocation": {"Bucket": bucket, "Key": manifest_key}}
        },
    )
    return response["Arn"]


@task
def create_quicksight_dataset(aws_account_id: int, dataset_name: str, data_source_arn: str) -> None:
    table_map = {
        "default": {
            "S3Source": {
                "DataSourceArn": data_source_arn,
                "InputColumns": [{"Name": name, "Type": "STRING"} for name in SAMPLE_DATA_COLUMNS],
            }
        }
    }

    boto3.client("quicksight").create_data_set(
        AwsAccountId=aws_account_id,
        DataSetId=dataset_name,
        Name=dataset_name,
        PhysicalTableMap=table_map,
        ImportMode="SPICE",
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_quicksight_data_source(aws_account_id: str, datasource_name: str):
    boto3.client("quicksight").delete_data_source(AwsAccountId=aws_account_id, DataSourceId=datasource_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_dataset(aws_account_id: str, dataset_name: str):
    boto3.client("quicksight").delete_data_set(AwsAccountId=aws_account_id, DataSetId=dataset_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_ingestion(aws_account_id: str, dataset_name: str, ingestion_name: str) -> None:
    client = boto3.client("quicksight")
    with contextlib.suppress(client.exceptions.ResourceNotFoundException):
        # suppress ResourceNotFoundException: Ingestion has already terminated on its own.
        client.cancel_ingestion(
            AwsAccountId=aws_account_id,
            DataSetId=dataset_name,
            IngestionId=ingestion_name,
        )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    account_id = get_aws_account_id()

    env_id = test_context[ENV_ID_KEY]
    bucket_name = f"{env_id}-quicksight-bucket"
    data_filename = "sample_data.csv"
    dataset_id = f"{env_id}-data-set"
    datasource_id = f"{env_id}-data-source"
    ingestion_id = f"{env_id}-ingestion"
    manifest_filename = f"{env_id}-manifest.json"
    manifest_contents = {"fileLocations": [{"URIs": [f"s3://{bucket_name}/{data_filename}"]}]}

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=bucket_name)

    upload_manifest_file = S3CreateObjectOperator(
        task_id="upload_manifest_file",
        s3_bucket=bucket_name,
        s3_key=manifest_filename,
        data=json.dumps(manifest_contents),
        replace=True,
    )

    upload_sample_data = S3CreateObjectOperator(
        task_id="upload_sample_data",
        s3_bucket=bucket_name,
        s3_key=data_filename,
        data=SAMPLE_DATA,
        replace=True,
    )

    data_source = create_quicksight_data_source(
        aws_account_id=account_id,
        datasource_name=datasource_id,
        bucket=bucket_name,
        manifest_key=manifest_filename,
    )

    create_dataset = create_quicksight_dataset(account_id, dataset_id, data_source)

    # [START howto_operator_quicksight_create_ingestion]
    create_ingestion = QuickSightCreateIngestionOperator(
        task_id="create_ingestion",
        data_set_id=dataset_id,
        ingestion_id=ingestion_id,
    )
    # [END howto_operator_quicksight_create_ingestion]

    # QuickSightCreateIngestionOperator waits by default, setting as False to test the Sensor below.
    create_ingestion.wait_for_completion = False

    # If this sensor appears to freeze with a "QUEUED" status, see note above.
    # [START howto_sensor_quicksight]
    await_job = QuickSightSensor(
        task_id="await_job",
        data_set_id=dataset_id,
        ingestion_id=ingestion_id,
    )
    # [END howto_sensor_quicksight]
    await_job.poke_interval = 10

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        account_id,
        create_s3_bucket,
        upload_manifest_file,
        upload_sample_data,
        data_source,
        create_dataset,
        # TEST BODY
        create_ingestion,
        await_job,
        # TEST TEARDOWN
        delete_dataset(account_id, dataset_id),
        delete_quicksight_data_source(account_id, datasource_id),
        delete_ingestion(account_id, dataset_id, ingestion_id),
        delete_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.providers.amazon.aws.operators.ec2 import (
    EC2CreateInstanceOperator,
    EC2HibernateInstanceOperator,
    EC2RebootInstanceOperator,
    EC2StartInstanceOperator,
    EC2StopInstanceOperator,
    EC2TerminateInstanceOperator,
)
from airflow.providers.amazon.aws.sensors.ec2 import EC2InstanceStateSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder
from system.amazon.aws.utils.ec2 import get_latest_ami_id

DAG_ID = "example_ec2"

sys_test_context_task = SystemTestContextBuilder().build()


@task
def create_key_pair(key_name: str):
    client = boto3.client("ec2")

    key_pair_id = client.create_key_pair(KeyName=key_name)["KeyName"]
    # Creating the key takes a very short but measurable time, preventing race condition:
    client.get_waiter("key_pair_exists").wait(KeyNames=[key_pair_id])

    return key_pair_id


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_key_pair(key_pair_id: str):
    boto3.client("ec2").delete_key_pair(KeyName=key_pair_id)


@task
def parse_response(instance_ids: list):
    return instance_ids[0]


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    instance_name = f"{env_id}-instance"
    key_name = create_key_pair(key_name=f"{env_id}_key_pair")
    image_id = get_latest_ami_id()

    config = {
        "InstanceType": "t4g.micro",
        "KeyName": key_name,
        "TagSpecifications": [
            {"ResourceType": "instance", "Tags": [{"Key": "Name", "Value": instance_name}]}
        ],
        # Use IMDSv2 for greater security, see the following doc for more details:
        # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/configuring-instance-metadata-service.html
        "MetadataOptions": {"HttpEndpoint": "enabled", "HttpTokens": "required"},
        "HibernationOptions": {"Configured": True},
        "BlockDeviceMappings": [
            {"DeviceName": "/dev/xvda", "Ebs": {"Encrypted": True, "DeleteOnTermination": True}}
        ],
    }

    # EC2CreateInstanceOperator creates and starts the EC2 instances. To test the EC2StartInstanceOperator,
    # we will stop the instance, then start them again before terminating them.

    # [START howto_operator_ec2_create_instance]
    create_instance = EC2CreateInstanceOperator(
        task_id="create_instance",
        image_id=image_id,
        max_count=1,
        min_count=1,
        config=config,
    )
    # [END howto_operator_ec2_create_instance]
    create_instance.wait_for_completion = True
    instance_id = parse_response(create_instance.output)
    # [START howto_operator_ec2_stop_instance]
    stop_instance = EC2StopInstanceOperator(
        task_id="stop_instance",
        instance_id=instance_id,
    )
    # [END howto_operator_ec2_stop_instance]
    stop_instance.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_operator_ec2_start_instance]
    start_instance = EC2StartInstanceOperator(
        task_id="start_instance",
        instance_id=instance_id,
    )
    # [END howto_operator_ec2_start_instance]

    # [START howto_sensor_ec2_instance_state]
    await_instance = EC2InstanceStateSensor(
        task_id="await_instance",
        instance_id=instance_id,
        target_state="running",
    )
    # [END howto_sensor_ec2_instance_state]

    # [START howto_operator_ec2_reboot_instance]
    reboot_instance = EC2RebootInstanceOperator(
        task_id="reboot_instace",
        instance_ids=instance_id,
    )
    # [END howto_operator_ec2_reboot_instance]
    reboot_instance.wait_for_completion = True

    # [START howto_operator_ec2_hibernate_instance]
    hibernate_instance = EC2HibernateInstanceOperator(
        task_id="hibernate_instace",
        instance_ids=instance_id,
    )
    # [END howto_operator_ec2_hibernate_instance]
    hibernate_instance.wait_for_completion = True
    hibernate_instance.poll_interval = 60
    hibernate_instance.max_attempts = 40

    # [START howto_operator_ec2_terminate_instance]
    terminate_instance = EC2TerminateInstanceOperator(
        task_id="terminate_instance",
        instance_ids=instance_id,
        wait_for_completion=True,
    )
    # [END howto_operator_ec2_terminate_instance]
    terminate_instance.trigger_rule = TriggerRule.ALL_DONE
    chain(
        # TEST SETUP
        test_context,
        key_name,
        image_id,
        # TEST BODY
        create_instance,
        instance_id,
        stop_instance,
        start_instance,
        await_instance,
        reboot_instance,
        hibernate_instance,
        terminate_instance,
        # TEST TEARDOWN
        delete_key_pair(key_name),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This is a more advanced example dag for using `GoogleApiToS3Operator` which uses xcom to pass data between
tasks to retrieve specific information about YouTube videos:

First it searches for up to 50 videos (due to pagination) in a given time range
(YOUTUBE_VIDEO_PUBLISHED_AFTER, YOUTUBE_VIDEO_PUBLISHED_BEFORE) on a YouTube channel (YOUTUBE_CHANNEL_ID)
saves the response in S3 + passes over the YouTube IDs to the next request which then gets the information
(YOUTUBE_VIDEO_FIELDS) for the requested videos and saves them in S3 (S3_DESTINATION_KEY).

Further information:

YOUTUBE_VIDEO_PUBLISHED_AFTER and YOUTUBE_VIDEO_PUBLISHED_BEFORE needs to be formatted
``YYYY-MM-DDThh:mm:ss.sZ``.
See https://developers.google.com/youtube/v3/docs/search/list for more information.
YOUTUBE_VIDEO_PARTS depends on the fields you pass via YOUTUBE_VIDEO_FIELDS. See
https://developers.google.com/youtube/v3/docs/videos/list#parameters for more information.
YOUTUBE_CONN_ID is optional for public videos. It does only need to authenticate when there are private videos
on a YouTube channel you want to retrieve.

Authentication:
In order for the DAG to run, the GoogleApiToS3Operator needs to authenticate to Google Cloud APIs.
Further information about authenticating can be found:
https://cloud.google.com/docs/authentication/getting-started
https://cloud.google.com/docs/authentication/provide-credentials-adc

https://airflow.apache.org/docs/apache-airflow-providers-google/stable/connections/gcp.html

The required scope for this DAG is https://www.googleapis.com/auth/youtube.readonly.
This can be set via the environment variable AIRFLOW_CONN_GOOGLE_CLOUD_DEFAULT,
or by creating a custom connection.
"""

from __future__ import annotations

import json
from datetime import datetime

import boto3

from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.transfers.google_api_to_s3 import GoogleApiToS3Operator

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder
from tests_common.test_utils.api_client_helpers import make_authenticated_rest_api_request

DAG_ID = "example_google_api_youtube_to_s3"

YOUTUBE_CHANNEL_ID = "UCSXwxpWZQ7XZ1WL3wqevChA"
YOUTUBE_VIDEO_PUBLISHED_AFTER = "2019-09-25T00:00:00Z"
YOUTUBE_VIDEO_PUBLISHED_BEFORE = "2019-10-18T00:00:00Z"
YOUTUBE_VIDEO_PARTS = "snippet"
YOUTUBE_VIDEO_FIELDS = "items(id,snippet(description,publishedAt,tags,title))"

SECRET_ARN_KEY = "SECRET_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(SECRET_ARN_KEY).build()


@task
def create_connection_gcp(conn_id_name: str, secret_arn: str):
    json_data = boto3.client("secretsmanager").get_secret_value(SecretId=secret_arn)["SecretString"]
    conn_extra = {
        "scope": "https://www.googleapis.com/auth/youtube.readonly",
        "project": "aws-oss-airflow",
        "keyfile_dict": json_data,
    }
    make_authenticated_rest_api_request(
        path="/api/v2/connections",
        method="POST",
        body={
            "connection_id": conn_id_name,
            "conn_type": "google_cloud_platform",
            "extra": json.dumps(conn_extra),
        },
    )


@task(task_id="wait_for_s3_bucket")
def wait_for_bucket(s3_bucket_name):
    waiter = boto3.client("s3").get_waiter("bucket_exists")
    waiter.wait(Bucket=s3_bucket_name)


@task(task_id="transform_video_ids")
def transform_video_ids(**kwargs):
    task_instance = kwargs["task_instance"]
    output = task_instance.xcom_pull(task_ids="video_ids_to_s3", key="video_ids_response")
    video_ids = [item["id"]["videoId"] for item in output["items"]]

    if not video_ids:
        video_ids = []

    kwargs["task_instance"].xcom_push(key="video_ids", value={"id": ",".join(video_ids)})


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),  # Override to match your needs
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    conn_id_name = f"{env_id}-conn-id"
    secret_arn = test_context[SECRET_ARN_KEY]

    set_up_connection = create_connection_gcp(conn_id_name, secret_arn=secret_arn)

    s3_bucket_name = f"{env_id}-bucket"

    create_s3_bucket = S3CreateBucketOperator(task_id="create-s3-bucket", bucket_name=s3_bucket_name)

    wait_for_bucket_creation = wait_for_bucket(s3_bucket_name=s3_bucket_name)

    # [START howto_transfer_google_api_youtube_search_to_s3]
    video_ids_to_s3 = GoogleApiToS3Operator(
        task_id="video_ids_to_s3",
        google_api_service_name="youtube",
        google_api_service_version="v3",
        google_api_endpoint_path="youtube.search.list",
        gcp_conn_id=conn_id_name,
        google_api_endpoint_params={
            "part": "snippet",
            "channelId": YOUTUBE_CHANNEL_ID,
            "maxResults": 50,
            "publishedAfter": YOUTUBE_VIDEO_PUBLISHED_AFTER,
            "publishedBefore": YOUTUBE_VIDEO_PUBLISHED_BEFORE,
            "type": "video",
            "fields": "items/id/videoId",
        },
        google_api_response_via_xcom="video_ids_response",
        s3_destination_key=f"https://s3.us-west-2.amazonaws.com/{s3_bucket_name}/youtube_search",
        s3_overwrite=True,
    )
    # [END howto_transfer_google_api_youtube_search_to_s3]

    transform_video_ids_task = transform_video_ids()

    # [START howto_transfer_google_api_youtube_list_to_s3]
    video_data_to_s3 = GoogleApiToS3Operator(
        task_id="video_data_to_s3",
        google_api_service_name="youtube",
        google_api_service_version="v3",
        gcp_conn_id=conn_id_name,
        google_api_endpoint_path="youtube.videos.list",
        google_api_endpoint_params={
            "part": YOUTUBE_VIDEO_PARTS,
            "maxResults": 50,
            "fields": YOUTUBE_VIDEO_FIELDS,
        },
        google_api_endpoint_params_via_xcom="video_ids",
        s3_destination_key=f"https://s3.us-west-2.amazonaws.com/{s3_bucket_name}/youtube_videos",
        s3_overwrite=True,
    )
    # [END howto_transfer_google_api_youtube_list_to_s3]

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=s3_bucket_name,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    chain(
        # TEST SETUP
        test_context,
        set_up_connection,
        create_s3_bucket,
        wait_for_bucket_creation,
        # TEST BODY
        video_ids_to_s3,
        transform_video_ids_task,
        video_data_to_s3,
        # TEST TEARDOWN
        delete_s3_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook


class StsHook(AwsBaseHook):
    """
    Interact with AWS Security Token Service (STS).

    Provide thin wrapper around :external+boto3:py:class:`boto3.client("sts") <STS.Client>`.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(client_type="sts", *args, **kwargs)

    def get_account_number(self) -> str:
        """
        Get the account Number.

        .. seealso::
            - :external+boto3:py:meth:`STS.Client.get_caller_identity`
        """
        try:
            return self.get_conn().get_caller_identity()["Account"]
        except Exception as general_error:
            self.log.error("Failed to get the AWS Account Number, error: %s", general_error)
            raise
